/*
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: acccnf.c,v 1.1 2008/07/04 17:30:05 ca Exp $")
#include "sm/error.h"
#include "sm/assert.h"

#include "sm/sm-conf.h"
#include "sm/sm-conf-prt.h"

#include "smar.h"
#include "log.h"

/*
**  SMAR_CNF_RHS -- Go through elements in the RHS of a "rcpt conf"
**
**	Parameters:
**		smar_ctx -- SMAR context
**		smar_addr -- address context
**
**	Returns:
**		usual sm_error code
**		if ok: smar_addr->ara_status should have result of lookup(s)
**
**	Locking: smar_addr must be under control of the caller.
**
**	Last code review:
**	Last code change:
*/

typedef struct smar_ss_rcpt_cnf_S	smar_ss_rcpt_cnf_T, *smar_ss_rcpt_cnf_P;
struct smar_ss_rcpt_cnf_S
{
	uint	asrc_flags;
};

#if 0
static sm_conf_definition_T
smar_tlsreq_defs[] = {
TLSREQ_DEFS(tlsreq_cnf_T, tlsreqcnf),

/* Sentinel */
{ SM_CONF_DEF_MAGIC, NULL, 0, 0, 0, NULL, 0, NULL, NULL, NULL, NULL SM_LC_NO_ISSET SM_LC_SET_MAGIC(0)}
};
#endif

sm_conf_definition_T const
ar_src_flag_names[]
=
{
{ SM_CONF_DEF_MAGIC,
	"greylisting",			sm_conf_type_choice_value,
	SMARA_LT_GREY_FULL,
	0, NULL, 0, NULL, NULL, NULL,
	"greylisting"
	SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0) },

/* Sentinel */
{ SM_CONF_DEF_MAGIC, NULL, 0, 0, 0, NULL, 0, NULL, NULL, NULL, NULL SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0)}
};

static sm_conf_definition_T
ar_ssrcnf_defs[] = {
#if 0
{ SM_CONF_DEF_MAGIC, "tls_requirements", sm_conf_type_section,
	offsetof(ss_sess_T, ssse_tlsreq_cnf), sizeof(tlsreq_cnf_T), NULL,
	SM_CONF_FLAG_KEEP_DEFAULT,
	ssse_tlsreq_defs,
	NULL, NULL, NULL
	SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0) },
#endif

{ SM_CONF_DEF_MAGIC, "flags",		sm_conf_type_choice,
	offsetof(smar_ss_rcpt_cnf_T, asrc_flags),	sizeof(uint32_t),
	0,
	SM_CONF_FLAG_MULTIPLE|SM_CONF_FLAG_OR|SM_CONF_FLAG_NUMERIC|SM_CONF_FLAG_PLURAL,
	ar_src_flag_names,
	NULL,	NULL,
	"flags to influence behavior per RCPT"
	SM_LC_NO_ISSET
	SM_LC_SET_MAGIC(0) },

/* Sentinel */
{ SM_CONF_DEF_MAGIC, NULL, 0, 0, 0, NULL, 0, NULL, NULL, NULL, NULL SM_LC_NO_ISSET SM_LC_SET_MAGIC(0)}
};

sm_ret_T
smar_cnf_rhs(smar_ctx_P smar_ctx, smar_addr_P smar_addr)
{
	sm_ret_T ret;
	int err;
	smar_ss_rcpt_cnf_T smar_ss_rcpt_cnf;
#define AR_RCPT_CNF "smtps_rcpt_conf"

	ret = SM_SUCCESS;
	smar_addr->ara_status = SMTP_R_CONT;

	smar_addr->ara_cnf = sm_conf_new(AR_RCPT_CNF);
	if (NULL == smar_addr->ara_cnf) {
		err = errno;

		sm_log_write(smar_ctx->smar_lctx,
			AR_LCAT_CONFIG, AR_LMOD_CONFIG,
			SM_LOG_ERR, 8,
			"sev=ERROR, func=smar_cnf_rhs, sm_conf_new=NULL, errno=%d\n",
			err);
		return sm_error_temp(SM_EM_AR, ENOMEM);
	}
	err = sm_conf_read_data(smar_addr->ara_cnf,
				(char *)sm_str_getdata(smar_addr->ara_rhs),
				sm_str_getlen(smar_addr->ara_rhs), true);
	if (err != 0) {
		sm_prt_conferr(AR_RCPT_CNF, smar_addr->ara_cnf, err, smioerr);
		sm_log_write(smar_ctx->smar_lctx,
			AR_LCAT_CONFIG, AR_LMOD_CONFIG,
			SM_LOG_ERR, 8,
			"sev=ERROR, func=smar_cnf_rhs, sm_conf_read_data=%m\n",
			err);
		sm_conf_destroy(smar_addr->ara_cnf);
		smar_addr->ara_cnf = NULL;
		return err;
	}

	err = sm_conf_scan(smar_addr->ara_cnf, ar_ssrcnf_defs, 0, &smar_ss_rcpt_cnf);
	if (err != 0) {
		sm_log_write(smar_ctx->smar_lctx,
			AR_LCAT_CONFIG, AR_LMOD_CONFIG,
			SM_LOG_ERR, 8,
			"sev=ERROR, func=smar_cnf_rhs, sm_conf_scan=%m\n",
			err);
#if 0
		/* has this been done? */
		sm_conf_destroy(smar_addr->ara_cnf);
		smar_addr->ara_cnf = NULL;
#endif
		return err;
	}

	/* set flags */
	SMAR_LEV_DPRINTF(3, (SMAR_DEBFP, "sev=DBG, func=smar_cnf_rhs, flags=%#x\n", smar_ss_rcpt_cnf.asrc_flags));
	smar_addr->ara_ltype |= smar_ss_rcpt_cnf.asrc_flags;

	return ret;
}
